#include "global.h"

static void on_destroy();

static void activate_pref();
void activate_results();
void activate_stop();


int main(int argc, char **argv) 
{
	g_thread_init (NULL);
  	gdk_threads_init ();
  	gdk_threads_enter ();
	
	GtkWidget *window1;
	gtk_init(&argc, &argv);
    
	glade_init();
	
	xml_main = glade_xml_new(FILENAME, ROOTNODE, NULL);
	glade_xml_signal_autoconnect (xml_main);
	xml_prefs = glade_xml_new(FILENAME, PREFNODE, NULL);
	glade_xml_signal_autoconnect (xml_prefs);
	xml_about = glade_xml_new(FILENAME, ABOUTNODE, NULL);
	glade_xml_signal_autoconnect (xml_about);
	xml_results = glade_xml_new(FILENAME, RESULTSNODE, NULL);
	glade_xml_signal_autoconnect (xml_results);
	
	//init_text_view();
	
	get_spec();

	window1 = glade_xml_get_widget (xml_main, "window_main");
	gtk_widget_show (window1);
    g_signal_connect (G_OBJECT (window1), "destroy", G_CALLBACK (on_destroy), NULL);

    if (!window1) {
        g_warning("something bad happened while creating the interface");
        return 1;
    }
    
    set_statusbar("Press 'Start' to start benchmark");
    
    done = TRUE;
    
	gtk_main();
	gdk_threads_leave ();

    return 0;

}

// ********************************************************************
// Called when the user wants to start the test
// ********************************************************************
void activate_start()
{
	GtkWidget *cbtn;
	cbtn = glade_xml_get_widget (xml_prefs, "cb_float");
	test_float = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cbtn));
	cbtn = glade_xml_get_widget (xml_prefs, "cb_integer");
	test_integer = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cbtn));
	
	run_suite();
}

// ********************************************************************
// If a benchmark is running, disable various bit and peices
// ********************************************************************
void set_busy(gboolean busy)
{
	//gdk_threads_enter ();
	GtkWidget *button;
	button = glade_xml_get_widget (xml_main, "tb_start");
	gtk_widget_set_sensitive(button, !busy);
	gtk_widget_queue_draw(button);
	button = glade_xml_get_widget (xml_main, "tb_stop");
	gtk_widget_set_sensitive(button, busy);
	gtk_widget_queue_draw(button);
	button = glade_xml_get_widget (xml_main, "tb_quit");
	gtk_widget_set_sensitive(button, !busy);
	gtk_widget_queue_draw(button);
	button = glade_xml_get_widget (xml_main, "tb_results");
	gtk_widget_set_sensitive(button, !busy);
	gtk_widget_queue_draw(button);
	button = glade_xml_get_widget (xml_main, "tb_prefs");
	gtk_widget_set_sensitive(button, !busy);
	gtk_widget_queue_draw(button);
	button = glade_xml_get_widget (xml_main, "tb_about");
	gtk_widget_set_sensitive(button, !busy);
	gtk_widget_queue_draw(button);
	GtkWidget *menu_item;
	menu_item = glade_xml_get_widget (xml_main, "item_start");
	gtk_widget_set_sensitive(menu_item, !busy);
	menu_item = glade_xml_get_widget (xml_main, "item_stop");
	gtk_widget_set_sensitive(menu_item, busy);
	menu_item = glade_xml_get_widget (xml_main, "item_quit");
	gtk_widget_set_sensitive(menu_item, !busy);
	menu_item = glade_xml_get_widget (xml_main, "item_prefs");
	gtk_widget_set_sensitive(menu_item, !busy);
	menu_item = glade_xml_get_widget (xml_main, "item_results");
	gtk_widget_set_sensitive(menu_item, !busy);
	menu_item = glade_xml_get_widget (xml_main, "item_about");
	gtk_widget_set_sensitive(menu_item, !busy);
	//gdk_threads_leave ();
/*	if (!busy)
	{
		// Not doing anything so make sure progress bar is at 0%
		GtkWidget *pb;
		pb = glade_xml_get_widget (xml_main, "pb_total");
		gdk_threads_enter ();
		gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pb), 0);
		gtk_widget_queue_draw(pb);
		gdk_threads_leave ();
	}
*/
}

// ********************************************************************
// Called when the user has selected 'About'
// ********************************************************************
void activate_about()
{
	GtkWidget *dialog;
	dialog = glade_xml_get_widget (xml_about, "dialog_about");
	gtk_about_dialog_set_name(GTK_ABOUT_DIALOG(dialog), PROGRAM_NAME );
	gtk_about_dialog_set_version(GTK_ABOUT_DIALOG(dialog), PROGRAM_VERSION );
	gtk_widget_show (dialog);

}

// ********************************************************************
// Called when the user has selected 'Results'
// ********************************************************************
void activate_results()
{
	GtkWidget *window;
	window = glade_xml_get_widget (xml_results, "window_results");
	gtk_widget_show (window);
}

// ********************************************************************
// Close the preferences window
// ********************************************************************
void cancel_about()
{
	GtkWidget *window;
	window = glade_xml_get_widget (xml_about, "dialog_about");
	gtk_widget_hide (window);
}

// ********************************************************************
// Helper function to set the progress bar to a given %
// ********************************************************************
void set_progress(gfloat progress)
{
	GtkWidget *pb;
	pb = glade_xml_get_widget (xml_main, "pb_test");
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pb), progress);
	gtk_widget_queue_draw(pb);
}

// ********************************************************************
// Helper function to set the 'total' progress bar to a given %
// ********************************************************************
void set_progress_total(gfloat progress)
{
	GtkWidget *pb;
	pb = glade_xml_get_widget (xml_main, "pb_total");
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pb), progress);
	gtk_widget_queue_draw(pb);
}

// ********************************************************************
// Helper function to display text on the statusbar
// ********************************************************************
void set_statusbar(gchar *text)
{
	GtkWidget *sb;
	sb = glade_xml_get_widget (xml_main, "statusbar1");
	guint sb_context = gtk_statusbar_get_context_id (GTK_STATUSBAR (sb), "???");
	gtk_statusbar_push (GTK_STATUSBAR (sb), sb_context, text);
}

// ********************************************************************
// Helper function to display text in the output box
// ********************************************************************
void set_test_name(gchar *text)
{
	GtkWidget *lbl;
	lbl = glade_xml_get_widget (xml_main, "lbl_testName");
	gtk_label_set_text(GTK_LABEL(lbl), text);
	gtk_widget_queue_draw(lbl);
}

// ********************************************************************
// Called when the user wants to quit
// ********************************************************************
void activate_quit()
{
	if (done)
		gtk_exit(0);
}

void activate_prefs()
{
	GtkWidget *window_prefs;
	window_prefs = glade_xml_get_widget (xml_prefs, "window_prefs");
	gtk_widget_show (window_prefs);
}	

void activate_stop()
{
}

void apply_prefs()
{
	GtkWidget *window_prefs;
	window_prefs = glade_xml_get_widget (xml_prefs, "window_prefs");
	gtk_widget_hide (window_prefs);
}

// ********************************************************************
// Activate master thread that queues benchmarks
// ********************************************************************
void run_suite()
{
	//GtkWidget *fc;
	//fc = glade_xml_get_widget (xml_prefs, "fc_filechooser");
	//strcpy(file_path, gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(fc)));
	
	pthread_t pt_master;
	
	if (pthread_create(&pt_master, NULL, sub_master, NULL) != 0)
		printf("ERROR: Could not create master thread\n");

}

// ********************************************************************
// Master thread- run all selected benchmarks
// ********************************************************************
void *sub_master(void *arg)
{
	gchar tmp_str[200];
	GtkWidget *widget;
	widget = glade_xml_get_widget (xml_prefs, "sb_threads");
	gint threads = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	widget = glade_xml_get_widget (xml_prefs, "sb_time");
	gint run_time = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	widget = glade_xml_get_widget (xml_prefs, "sb_reps");
	gint reps = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));		
	
	gdk_threads_enter ();
	set_busy(TRUE);
	gdk_threads_leave ();
	
	fully_done = FALSE;
	start_timer_1();
	pthread_t pt_monitor_total;
	if (pthread_create(&pt_monitor_total, NULL, sub_monitor_total, NULL) != 0)
		printf("ERROR: Could not create monitor thread\n");
	
	float result = 0.0f;
	float *results_single;
	results_single = (float*)malloc(sizeof(float)*reps);
	float *results_multi;
	results_multi = (float*)malloc(sizeof(float)*reps);
	
	float weighted_result, mean;
	float result_single_int, result_single_float, result_multi_int, result_multi_float;
	float sd_single_int, sd_single_float, sd_multi_int, sd_multi_float;
	float peak_single_int, peak_single_float, peak_multi_int, peak_multi_float;
	float low_single_int, low_single_float, low_multi_int, low_multi_float;
	
	if (test_integer)
	{
		int i=0;
		float total_multi=0.0f;
		float total_single=0.0f;
		for (i=0; i<reps; i++)
		{	
			gdk_threads_enter ();
			widget = glade_xml_get_widget (xml_main, "l_runNumber");
			sprintf(tmp_str, "Run: %d/%d", i+1, reps);
			gtk_label_set_label(GTK_LABEL(widget), tmp_str);
			gdk_threads_leave ();
	
			result = run_integer(1, run_time);
			results_single[i]=result;
			
			if (threads>1)
			{
				result = run_integer(threads, run_time);
				results_multi[i]=result;
			}
		}
		
		// Work out the standard deviation for single & multi results
		mean = get_mean(results_single, reps);
		sd_single_int = get_sd(results_single, reps);
		peak_single_int = get_peak(results_single, reps);
		low_single_int = get_low(results_single, reps);
		result_single_int=mean;
		if (threads>1)
		{
			mean = get_mean(results_multi, reps);
			sd_multi_int = get_sd(results_multi, reps);
			peak_multi_int = get_peak(results_multi, reps);
			low_multi_int = get_low(results_multi, reps);
			result_multi_int=mean;
		}
	
	}
	
	if (test_float)
	{
		int i=0;
		float total_multi=0.0f;
		float total_single=0.0f;
		for (i=0; i<reps; i++)
		{	
			gdk_threads_enter ();
			widget = glade_xml_get_widget (xml_main, "l_runNumber");
			sprintf(tmp_str, "Run: %d/%d", i+1, reps);
			gtk_label_set_label(GTK_LABEL(widget), tmp_str);
			gdk_threads_leave ();
	
			result = run_float(1, run_time);
			results_single[i]=result;
			
			if (threads>1)
			{
				result = run_float(threads, run_time);
				results_multi[i]=result;
			}
		}
		
		
		// Work out the standard deviation for single & multi results
		mean = get_mean(results_single, reps);
		sd_single_float = get_sd(results_single, reps);
		peak_single_float = get_peak(results_single, reps);
		low_single_float = get_low(results_single, reps);
		result_single_float = mean;
		if (threads>1)
		{
			mean = get_mean(results_multi, reps);
			sd_multi_float = get_sd(results_multi, reps);
			peak_multi_float = get_peak(results_multi, reps);
			low_multi_float = get_low(results_multi, reps);
			result_multi_float = mean;
		}
	
	}
	
	weighted_result = 4 * (result_single_int + result_multi_int + result_single_float + result_multi_float);
	weighted_result += 2 * (peak_single_int   + peak_multi_int   + peak_single_float   + peak_multi_float);
	weighted_result += 1* (low_single_int    + low_multi_int    + low_single_float    + low_multi_float);
	weighted_result = weighted_result / 8;
	
	
	fully_done = TRUE;

	gdk_threads_enter ();
	set_busy(FALSE);
	gdk_threads_leave ();
	set_progress(0);
	set_progress_total(0);
	
	gdk_threads_enter ();
	set_test_name("-");
	gdk_threads_leave ();
	
	gdk_threads_enter ();
	widget = glade_xml_get_widget (xml_main, "l_runNumber");
	gtk_label_set_label(GTK_LABEL(widget), "Run: -");
	widget = glade_xml_get_widget (xml_main, "l_eta");
	gtk_label_set_label(GTK_LABEL(widget), "ETA: -");
	
	// Update results tab for CPU
	widget = glade_xml_get_widget (xml_results, "e_threads");
	sprintf(tmp_str, "%d", threads);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_results");
	sprintf(tmp_str, "%.0f", weighted_result);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);

	widget = glade_xml_get_widget (xml_results, "e_cpu_intsingle_mean");
	sprintf(tmp_str, "%.0f", result_single_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intsingle_sd");
	sprintf(tmp_str, "%.0f", sd_single_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intsingle_peak");
	sprintf(tmp_str, "%.0f", peak_single_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intsingle_low");
	sprintf(tmp_str, "%.0f", low_single_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatsingle_mean");
	sprintf(tmp_str, "%.0f", result_single_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatsingle_sd");
	sprintf(tmp_str, "%.0f", sd_single_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatsingle_peak");
	sprintf(tmp_str, "%.0f", peak_single_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatsingle_low");
	sprintf(tmp_str, "%.0f", low_single_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	
	widget = glade_xml_get_widget (xml_results, "e_cpu_intmulti_mean");
	sprintf(tmp_str, "%.0f", result_multi_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intmulti_sd");
	sprintf(tmp_str, "%.0f", sd_multi_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intmulti_peak");
	sprintf(tmp_str, "%.0f", peak_multi_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_intmulti_low");
	sprintf(tmp_str, "%.0f", low_multi_int);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatmulti_mean");
	sprintf(tmp_str, "%.0f", result_multi_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatmulti_sd");
	sprintf(tmp_str, "%.0f", sd_multi_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatmulti_peak");
	sprintf(tmp_str, "%.0f", peak_multi_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	widget = glade_xml_get_widget (xml_results, "e_cpu_floatmulti_low");
	sprintf(tmp_str, "%.0f", low_multi_float);
	gtk_entry_set_text(GTK_ENTRY(widget), tmp_str);
	
	gdk_threads_leave ();

}

float get_sd(float *values, int num)
{
	float mean=get_mean(values, num);
	
	float sd=0;
	int i;
	for (i=0; i<num; i++)
	{
		sd+=(values[i]-mean)*(values[i]-mean);
		//printf("V(%2d)=%3.1f   D=%3.1f\n",i,values[i],sd);
	}
	
	sd = sqrt(sd/num);
		
	return sd;
}

float get_mean(float *values, int num)
{
	float total=0;
	int i;
	for (i=0; i<num; i++)
		total+=values[i];
	return total/num;
}

float get_peak(float *values, int num)
{
	int i;
	float peak=-1;
	for (i=0; i<num; i++)
		if (values[i]>peak) peak=values[i];
	return peak;
}

float get_low(float *values, int num)
{
	int i;
	float low=9999999.0f;
	for (i=0; i<num; i++)
		if (values[i]<low) low=values[i];
	return low;
}

// ********************************************************************
// Monitor thread- used to update the 'total' progess bar
// ********************************************************************
void *sub_monitor_total(void *arg)
{
	gchar tmp_str[200];
	GtkWidget *spinbox;
	spinbox = glade_xml_get_widget (xml_prefs, "sb_time");
	gint run_time = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbox));
	spinbox = glade_xml_get_widget (xml_prefs, "sb_reps");
	gint run_reps = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbox));
	spinbox = glade_xml_get_widget (xml_prefs, "sb_threads");
	gint run_threads = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbox));
	gint run_time_total;
	if (run_threads>1)
		run_time_total = (run_time * run_reps * 2);
	else
		run_time_total = (run_time * run_reps);
	if (test_integer && test_float)
		run_time_total *= 2;
	
	while(!fully_done)
	{
		long elapsed = get_time_1();
		int secs = (float)elapsed/10000000;
		float prog_total = (float)secs/run_time_total;
		if (prog_total>1) prog_total=1.0f;
		if (prog_total<0) prog_total=0.0f;
		gdk_threads_enter ();
		set_progress_total(prog_total);
		gdk_threads_leave ();
		
		gdk_threads_enter ();
		GtkWidget *spinbox;
		spinbox = glade_xml_get_widget (xml_main, "l_eta");
		if (run_time_total-secs>60)
		{
			sprintf(tmp_str,"ETA: %d minutes",(run_time_total-secs)/60);
		}
		else
		{
			sprintf(tmp_str,"ETA: %d seconds",run_time_total-secs);
		}

		gtk_label_set_label(GTK_LABEL(spinbox),tmp_str);
		gdk_threads_leave ();
		
		
		sleep(1);
	}
}


// ********************************************************************
// Monitor thread- used to update the progess bar
// ********************************************************************
void *sub_monitor(void *arg)
{
	GtkWidget *spinbox;
	spinbox = glade_xml_get_widget (xml_prefs, "sb_time");
	gint run_time = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbox));
	
	while(!done)
	{
		long elapsed = get_time_2();
		int secs = (float)elapsed/10000000;
		float prog = (float)secs/run_time;
		if (prog>1) prog=1.0f;
		if (prog<0) prog=0.0f;
		gdk_threads_enter ();
		set_progress(prog);
		gdk_threads_leave ();
		sleep(1);
	}
}

void test_destroy()
{
	// TODO: Disable running threads cleanly
}

// ********************************************************
// Exit the application
// ********************************************************
static void on_destroy (GtkWidget * widget, gpointer data)
{
     gtk_main_quit();
}
